// ----------------------------------------------------------------------------
// Project:	Implementation Alternatives for Bottom-Up Evaluation
// Filename:	test5mat.cpp
// Purpose:	Time Measurements for Rule Implementations
// Last Change:	06.02.2010
// Language:	C++
// EMail:	brass@acm.org
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Univ. Halle, Von-Seckendorff-Platz 1, D-06120 Halle, GERMANY
// Copyright:	(c) 2010 by Stefan Brass
//		Downloads and modifications for tests and scientific purposes
//		permitted, provided that some acknowledgement of my work
//		remains, and modified versions are clearly marked as such.
//		There is no warranty at all - this code may contain bugs.
// ----------------------------------------------------------------------------

// Example Program 5:
//
// p1(A) :- q(A).
// p2(A,1) :- p1(A).
// p2(A,2) :- p1(A).
// p3(A,B,1) :- p2(A,B).
// p3(A,B,2) :- p2(A,B).
// p4(A,B,C,1) :- p3(A,B,C).
// p4(A,B,C,2) :- p3(A,B,C).
// p5(A,B,C,D,1) :- p4(A,B,C,D).
// p5(A,B,C,D,2) :- p4(A,B,C,D).
// p6(A,B,C,D,E,1) :- p5(A,B,C,D,E).
// p6(A,B,C,D,E,2) :- p5(A,B,C,D,E).
// p7(A,B,C,D,E,F,1) :- p6(A,B,C,D,E,F).
// p7(A,B,C,D,E,F,2) :- p6(A,B,C,D,E,F).

// q := {(i)|i = 1,...,50000}


#include <iostream>
using namespace std;

#if defined(_MSC_VER) || defined(_WINDOWS_)
#define WINDOWS
#else
#define UNIX
#endif

// ----------------------------------------------------------------------------
// Sizes for test:
// ----------------------------------------------------------------------------

const int Q_TUPLES = 50000;

const bool PRINT_EXTENSION = false;

// ----------------------------------------------------------------------------
// Timing Functions (UNIX Version):
// ----------------------------------------------------------------------------

#ifdef UNIX
#include <sys/time.h>
#include <time.h>

class timer {
public:
	void start()
	{
		gettimeofday(&Time_Start, 0);
		Clock_Start = clock();
	}

	void end()
	{
		gettimeofday(&Time_End, 0);
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		long sec = Time_End.tv_sec - Time_Start.tv_sec;
		long usec = Time_End.tv_usec - Time_Start.tv_usec;
		if(usec < 0) {
			usec += 1000000;
			sec -= 1;
		}
		return sec * 1000 + (usec + 500) / 1000;
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	timeval Time_Start, Time_End;
	clock_t Clock_Start, Clock_End;
};
#endif

// ----------------------------------------------------------------------------
// Timing Functions (Windows Version):
// ----------------------------------------------------------------------------

#ifdef WINDOWS
#include <windows.h>
#include <time.h>

class timer {
public:
	void start()
	{
		GetSystemTimeAsFileTime(&(Start.ft));
		Clock_Start = clock();
	}

	void end()
	{
		GetSystemTimeAsFileTime(&(End.ft));
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		return static_cast<long> ((End.ns100 - Start.ns100) / 10000);
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	union {
		long long ns100;
		FILETIME ft;
	} Start, End;
	clock_t Clock_Start, Clock_End;
};
#endif


// ----------------------------------------------------------------------------
// Predicate p1 - implemented as linked list
// ----------------------------------------------------------------------------

class p1 {
	friend class p1_cursor;
public:
	int col_1() const {return Col_1; }

	static void insert(int c1) {
		p1 *new_elem = new p1(c1);
		if(Last)
			Last->Next = new_elem;
		else
			First = new_elem;
		Last = new_elem;
	}

private:
	p1(int c1)
	{
		Col_1 = c1;
		Next = 0;
	}

	int Col_1;
	p1 *Next;

	static class p1* First;
	static class p1* Last;
};

class p1* p1::First;
class p1* p1::Last;

class p1_cursor {
public:
	p1_cursor()
	{
		Before_First = true;
		Current = 0;
	}

	void open()
	{
		Before_First = true;
	}

	bool fetch()
	{
		if(Before_First) {
			Current = p1::First;
			Before_First = false;
		}
		else
			Current = Current->Next;
		return(Current != NULL);
	}

	int col_1() const { return Current->col_1(); }

	void close() {}

private:
	bool Before_First;
	p1*   Current;
};


// ----------------------------------------------------------------------------
// Predicate p2 - implemented as linked list
// ----------------------------------------------------------------------------

class p2 {
	friend class p2_cursor;
public:
	int col_1() const {return Col_1; }
	int col_2() const {return Col_2; }

	static void insert(int c1, int c2) {
		p2 *new_elem = new p2(c1, c2);
		if(Last)
			Last->Next = new_elem;
		else
			First = new_elem;
		Last = new_elem;
	}

private:
	p2(int c1, int c2)
	{
		Col_1 = c1;
		Col_2 = c2;
		Next = 0;
	}

	int Col_1;
	int Col_2;
	p2 *Next;

	static class p2* First;
	static class p2* Last;
};

class p2* p2::First;
class p2* p2::Last;

class p2_cursor {
public:
	p2_cursor()
	{
		Before_First = true;
		Current = 0;
	}

	void open()
	{
		Before_First = true;
	}

	bool fetch()
	{
		if(Before_First) {
			Current = p2::First;
			Before_First = false;
		}
		else
			Current = Current->Next;
		return(Current != NULL);
	}

	int col_1() const { return Current->col_1(); }
	int col_2() const { return Current->col_2(); }

	void close() {}

private:
	bool Before_First;
	p2*   Current;
};


// ----------------------------------------------------------------------------
// Predicate p3 - implemented as linked list
// ----------------------------------------------------------------------------

class p3 {
	friend class p3_cursor;
public:
	int col_1() const {return Col_1; }
	int col_2() const {return Col_2; }
	int col_3() const {return Col_3; }

	static void insert(int c1, int c2, int c3) {
		p3 *new_elem = new p3(c1, c2, c3);
		if(Last)
			Last->Next = new_elem;
		else
			First = new_elem;
		Last = new_elem;
	}

private:
	p3(int c1, int c2, int c3)
	{
		Col_1 = c1;
		Col_2 = c2;
		Col_3 = c3;
		Next = 0;
	}

	int Col_1;
	int Col_2;
	int Col_3;
	p3 *Next;

	static class p3* First;
	static class p3* Last;
};

class p3* p3::First;
class p3* p3::Last;

class p3_cursor {
public:
	p3_cursor()
	{
		Before_First = true;
		Current = 0;
	}

	void open()
	{
		Before_First = true;
	}

	bool fetch()
	{
		if(Before_First) {
			Current = p3::First;
			Before_First = false;
		}
		else
			Current = Current->Next;
		return(Current != NULL);
	}

	int col_1() const { return Current->col_1(); }
	int col_2() const { return Current->col_2(); }
	int col_3() const { return Current->col_3(); }

	void close() {}

private:
	bool Before_First;
	p3*   Current;
};

// ----------------------------------------------------------------------------
// Predicate p4 - implemented as linked list
// ----------------------------------------------------------------------------

class p4 {
	friend class p4_cursor;
public:
	int col_1() const {return Col_1; }
	int col_2() const {return Col_2; }
	int col_3() const {return Col_3; }
	int col_4() const {return Col_4; }

	static void insert(int c1, int c2, int c3, int c4) {
		p4 *new_elem = new p4(c1, c2, c3, c4);
		if(Last)
			Last->Next = new_elem;
		else
			First = new_elem;
		Last = new_elem;
	}

private:
	p4(int c1, int c2, int c3, int c4)
	{
		Col_1 = c1;
		Col_2 = c2;
		Col_3 = c3;
		Col_4 = c4;
		Next = 0;
	}

	int Col_1;
	int Col_2;
	int Col_3;
	int Col_4;
	p4 *Next;

	static class p4* First;
	static class p4* Last;
};

class p4* p4::First;
class p4* p4::Last;

class p4_cursor {
public:
	p4_cursor()
	{
		Before_First = true;
		Current = 0;
	}

	void open()
	{
		Before_First = true;
	}

	bool fetch()
	{
		if(Before_First) {
			Current = p4::First;
			Before_First = false;
		}
		else
			Current = Current->Next;
		return(Current != NULL);
	}

	int col_1() const { return Current->col_1(); }
	int col_2() const { return Current->col_2(); }
	int col_3() const { return Current->col_3(); }
	int col_4() const { return Current->col_4(); }

	void close() {}

private:
	bool Before_First;
	p4*   Current;
};

// ----------------------------------------------------------------------------
// Predicate p5 - implemented as linked list
// ----------------------------------------------------------------------------

class p5 {
	friend class p5_cursor;
public:
	int col_1() const {return Col_1; }
	int col_2() const {return Col_2; }
	int col_3() const {return Col_3; }
	int col_4() const {return Col_4; }
	int col_5() const {return Col_5; }

	static void insert(int c1, int c2, int c3, int c4, int c5) {
		p5 *new_elem = new p5(c1, c2, c3, c4, c5);
		if(Last)
			Last->Next = new_elem;
		else
			First = new_elem;
		Last = new_elem;
	}

private:
	p5(int c1, int c2, int c3, int c4, int c5)
	{
		Col_1 = c1;
		Col_2 = c2;
		Col_3 = c3;
		Col_4 = c4;
		Col_5 = c5;
		Next = 0;
	}

	int Col_1;
	int Col_2;
	int Col_3;
	int Col_4;
	int Col_5;
	p5 *Next;

	static class p5* First;
	static class p5* Last;
};

class p5* p5::First;
class p5* p5::Last;

class p5_cursor {
public:
	p5_cursor()
	{
		Before_First = true;
		Current = 0;
	}

	void open()
	{
		Before_First = true;
	}

	bool fetch()
	{
		if(Before_First) {
			Current = p5::First;
			Before_First = false;
		}
		else
			Current = Current->Next;
		return(Current != NULL);
	}

	int col_1() const { return Current->col_1(); }
	int col_2() const { return Current->col_2(); }
	int col_3() const { return Current->col_3(); }
	int col_4() const { return Current->col_4(); }
	int col_5() const { return Current->col_5(); }

	void close() {}

private:
	bool Before_First;
	p5*   Current;
};

// ----------------------------------------------------------------------------
// Predicate p6 - implemented as linked list
// ----------------------------------------------------------------------------

class p6 {
	friend class p6_cursor;
public:
	int col_1() const {return Col_1; }
	int col_2() const {return Col_2; }
	int col_3() const {return Col_3; }
	int col_4() const {return Col_4; }
	int col_5() const {return Col_5; }
	int col_6() const {return Col_6; }

	static void insert(int c1, int c2, int c3, int c4, int c5, int c6) {
		p6 *new_elem = new p6(c1, c2, c3, c4, c5, c6);
		if(Last)
			Last->Next = new_elem;
		else
			First = new_elem;
		Last = new_elem;
	}

private:
	p6(int c1, int c2, int c3, int c4, int c5, int c6)
	{
		Col_1 = c1;
		Col_2 = c2;
		Col_3 = c3;
		Col_4 = c4;
		Col_5 = c5;
		Col_6 = c6;
		Next = 0;
	}

	int Col_1;
	int Col_2;
	int Col_3;
	int Col_4;
	int Col_5;
	int Col_6;
	p6 *Next;

	static class p6* First;
	static class p6* Last;
};

class p6* p6::First;
class p6* p6::Last;

class p6_cursor {
public:
	p6_cursor()
	{
		Before_First = true;
		Current = 0;
	}

	void open()
	{
		Before_First = true;
	}

	bool fetch()
	{
		if(Before_First) {
			Current = p6::First;
			Before_First = false;
		}
		else
			Current = Current->Next;
		return(Current != NULL);
	}

	int col_1() const { return Current->col_1(); }
	int col_2() const { return Current->col_2(); }
	int col_3() const { return Current->col_3(); }
	int col_4() const { return Current->col_4(); }
	int col_5() const { return Current->col_5(); }
	int col_6() const { return Current->col_6(); }

	void close() {}

private:
	bool Before_First;
	p6*   Current;
};

// ----------------------------------------------------------------------------
// Predicate p7 - implemented as linked list
// ----------------------------------------------------------------------------

class p7 {
	friend class p7_cursor;
public:
	int col_1() const {return Col_1; }
	int col_2() const {return Col_2; }
	int col_3() const {return Col_3; }
	int col_4() const {return Col_4; }
	int col_5() const {return Col_5; }
	int col_6() const {return Col_6; }
	int col_7() const {return Col_7; }

	static void insert(int c1, int c2, int c3, int c4, int c5, int c6,
				int c7) {
		p7 *new_elem = new p7(c1, c2, c3, c4, c5, c6, c7);
		if(Last)
			Last->Next = new_elem;
		else
			First = new_elem;
		Last = new_elem;
	}

private:
	p7(int c1, int c2, int c3, int c4, int c5, int c6, int c7)
	{
		Col_1 = c1;
		Col_2 = c2;
		Col_3 = c3;
		Col_4 = c4;
		Col_5 = c5;
		Col_6 = c6;
		Col_7 = c7;
		Next = 0;
	}

	int Col_1;
	int Col_2;
	int Col_3;
	int Col_4;
	int Col_5;
	int Col_6;
	int Col_7;
	p7 *Next;

	static class p7* First;
	static class p7* Last;
};

class p7* p7::First;
class p7* p7::Last;

class p7_cursor {
public:
	p7_cursor()
	{
		Before_First = true;
		Current = 0;
	}

	void open()
	{
		Before_First = true;
	}

	bool fetch()
	{
		if(Before_First) {
			Current = p7::First;
			Before_First = false;
		}
		else
			Current = Current->Next;
		return(Current != NULL);
	}

	int col_1() const { return Current->col_1(); }
	int col_2() const { return Current->col_2(); }
	int col_3() const { return Current->col_3(); }
	int col_4() const { return Current->col_4(); }
	int col_5() const { return Current->col_5(); }
	int col_6() const { return Current->col_6(); }
	int col_7() const { return Current->col_7(); }

	void close() {}

private:
	bool Before_First;
	p7*   Current;
};

//-----------------------------------------------------------------------------
// Predicate q: Tuples generated by computation
//-----------------------------------------------------------------------------

class q_cursor {
public:
	q_cursor() { }

	void open() { i = 0; }

	bool fetch()
	{
		if(i < Q_TUPLES) {
			i++;
			return true;
		}
		else
			return false;
	}

	int col_1() const { return i; }

	void close() {}

private:
	int i;
};


//-----------------------------------------------------------------------------
// Rule Implementation: Materialization
//-----------------------------------------------------------------------------

void exec_prog()
{
	// p1(A) :- q(A).
	{
		q_cursor c;
		c.open();
		while(c.fetch()) {
			int A = c.col_1();
			p1::insert(A);
		}
		c.close();
	}

	// p2(A,1) :- p1(A).
	{
		p1_cursor c;
		c.open();
		while(c.fetch()) {
			int A = c.col_1();
			p2::insert(A,1);
		}
		c.close();
	}

	// p2(A,2) :- p1(A).
	{
		p1_cursor c;
		c.open();
		while(c.fetch()) {
			int A = c.col_1();
			p2::insert(A,2);
		}
		c.close();
	}

	// p3(A,B,1) :- p2(A,B).
	{
		p2_cursor c;
		c.open();
		while(c.fetch()) {
			int A = c.col_1();
			int B = c.col_2();
			p3::insert(A,B,1);
		}
		c.close();
	}

	// p3(A,B,2) :- p2(A,B).
	{
		p2_cursor c;
		c.open();
		while(c.fetch()) {
			int A = c.col_1();
			int B = c.col_2();
			p3::insert(A,B,2);
		}
		c.close();
	}

	// p4(A,B,C,1) :- p3(A,B,C).
	{
		p3_cursor c;
		c.open();
		while(c.fetch()) {
			int A = c.col_1();
			int B = c.col_2();
			int C = c.col_3();
			p4::insert(A,B,C,1);
		}
		c.close();
	}

	// p4(A,B,C,2) :- p3(A,B,C).
	{
		p3_cursor c;
		c.open();
		while(c.fetch()) {
			int A = c.col_1();
			int B = c.col_2();
			int C = c.col_3();
			p4::insert(A,B,C,2);
		}
		c.close();
	}

	// p5(A,B,C,D,1) :- p4(A,B,C,D).
	{
		p4_cursor c;
		c.open();
		while(c.fetch()) {
			int A = c.col_1();
			int B = c.col_2();
			int C = c.col_3();
			int D = c.col_4();
			p5::insert(A,B,C,D,1);
		}
		c.close();
	}

	// p5(A,B,C,D,2) :- p4(A,B,C,D).
	{
		p4_cursor c;
		c.open();
		while(c.fetch()) {
			int A = c.col_1();
			int B = c.col_2();
			int C = c.col_3();
			int D = c.col_4();
			p5::insert(A,B,C,D,2);
		}
		c.close();
	}

	// p6(A,B,C,D,E,1) :- p5(A,B,C,D,E).
	{
		p5_cursor c;
		c.open();
		while(c.fetch()) {
			int A = c.col_1();
			int B = c.col_2();
			int C = c.col_3();
			int D = c.col_4();
			int E = c.col_5();
			p6::insert(A,B,C,D,E,1);
		}
		c.close();
	}

	// p6(A,B,C,D,E,2) :- p5(A,B,C,D,E).
	{
		p5_cursor c;
		c.open();
		while(c.fetch()) {
			int A = c.col_1();
			int B = c.col_2();
			int C = c.col_3();
			int D = c.col_4();
			int E = c.col_5();
			p6::insert(A,B,C,D,E,2);
		}
		c.close();
	}

	// p7(A,B,C,D,E,F,1) :- p6(A,B,C,D,E,F).
	{
		p6_cursor c;
		c.open();
		while(c.fetch()) {
			int A = c.col_1();
			int B = c.col_2();
			int C = c.col_3();
			int D = c.col_4();
			int E = c.col_5();
			int F = c.col_6();
			p7::insert(A,B,C,D,E,F,1);
		}
		c.close();
	}

	// p7(A,B,C,D,E,F,2) :- p6(A,B,C,D,E,F).
	{
		p6_cursor c;
		c.open();
		while(c.fetch()) {
			int A = c.col_1();
			int B = c.col_2();
			int C = c.col_3();
			int D = c.col_4();
			int E = c.col_5();
			int F = c.col_6();
			p7::insert(A,B,C,D,E,F,2);
		}
		c.close();
	}
}

//-----------------------------------------------------------------------------
// Main Program:
//-----------------------------------------------------------------------------

int main()
{
	timer t;
	int c1_must_be = 1;
	int c1 = 0;
	int c2 = 0;
	int c3 = 0;
	int c4 = 0;
	int c5 = 0;
	int c6 = 0;
	int c7 = 0;

	cout << "Test of Method 1: Materialization [Program 5]\n";
	t.start();
	exec_prog();
	p7_cursor c;
	c.open();
	int num_tuples = 0;
	while(c.fetch()) {
		c1 = c.col_1();
		if(c1 != c1_must_be) {
			cout << "Error: Wrong value for c1:" << c.col_1()
				<< ", must be: " << c1_must_be << "!\n";
			exit(1);
		}
		if(c1_must_be == Q_TUPLES) {
			c1_must_be = 1;
		}
		else
			c1_must_be++;
		c2 += c.col_2();
		c3 += c.col_3();
		c4 += c.col_4();
		c5 += c.col_5();
		c6 += c.col_6();
		c7 += c.col_7();
		num_tuples++;
	}
	c.close();
	t.end();

	if(PRINT_EXTENSION) {
		cout << "Extension of p6:\n\n";
		c.open();
		while(c.fetch()) {
			cout << "p7(" << c.col_1() << ", " << c.col_2()
				<< ", " << c.col_3() << ", " << c.col_4()
				<< ", " << c.col_5() << ", " << c.col_6()
				<< ", " << c.col_7() << ").\n";
		}
		c.close();
		cout << "\n";
	}

	if(c2 != Q_TUPLES * 96 )
		cout << "Error: Wrong column sum c2: " << c2 << "\n";
	if(c3 != c2 || c4 != c2 || c5 != c2 || c6 != c2)
		cout << "Error: Wrong column sum c3, c4, c5, c6!\n";
	cout << "Total number of tuples: " << num_tuples << "\n";
	if(num_tuples != Q_TUPLES * 64)
		cout << "Error: Wrong number of tuples!\n";
	cout << "\nUsed time: " << t.realtime_msec() << " ms,  "
		<< "CPU: " << t.cputime_msec() << " ms.\n\n\n";

	return 0;
}


